from dataclasses import dataclass, field

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from diffusers import DDIMScheduler, DDPMScheduler, StableDiffusionPipeline, DDIMInverseScheduler
from diffusers.utils.import_utils import is_xformers_available
from tqdm import tqdm


import threestudio
from threestudio.models.prompt_processors.base import PromptProcessorOutput
from threestudio.utils.base import BaseObject
from threestudio.utils.misc import C, cleanup, parse_version
from threestudio.utils.ops import perpendicular_component
from threestudio.utils.typing import *


@threestudio.register("ours-inversion-guidance")
class OursInversionGuidance(BaseObject):
    @dataclass
    class Config(BaseObject.Config):
        pretrained_model_name_or_path: str = "runwayml/stable-diffusion-v1-5"
        enable_memory_efficient_attention: bool = False
        enable_sequential_cpu_offload: bool = False
        enable_attention_slicing: bool = False
        enable_channels_last_format: bool = False
        reset_t_every: int = -1
        guidance_scale: float = 100.0
        grad_clip: Optional[
            Any
        ] = None  # field(default_factory=lambda: [0, 2.0, 8.0, 1000])
        half_precision_weights: bool = True
        
        use_legacy_scheduler: bool = True

        min_step_percent: float = 0.02
        max_step_percent: float = 0.98
        sqrt_anneal: bool = True  # sqrt anneal proposed in HiFA: https://hifa-team.github.io/HiFA-site/
        trainer_max_steps: int = 10000
        use_img_loss: bool = False  # image-space SDS proposed in HiFA: https://hifa-team.github.io/HiFA-site/

        use_sjc: bool = False
        var_red: bool = True
        weighting_strategy: str = "sds"

        token_merging: bool = False
        token_merging_params: Optional[dict] = field(default_factory=dict)

        view_dependent_prompting: bool = True

        """Maximum number of batch items to evaluate guidance for (for debugging) and to save on disk. -1 means save all items."""
        max_items_eval: int = 4
        
        # Ours
        n_ddim_timesteps: int = 999
        t_map_power: float = 1.0
        inverse_guidance_scale: float = -10.0
        ddim_inverse_n_steps: int = 30
        inversion_t_threshold: int = 1000
        inversion_eta: float = 0.0
        linear_inversion_entropy: bool = False
        inversion_entropy_power: float = 1.0
        use_sds_scaling: bool = False
        back_cfg_scale: float = 1.0
        prompt_augmentation_cfg_scale: float = 1.0
        loss_reduction: str = "sum"
        noise_strategy: str = "inversion"
        
    cfg: Config

    def configure(self) -> None:
        threestudio.info(f"Loading Stable Diffusion ...")

        self.weights_dtype = (
            torch.float16 if self.cfg.half_precision_weights else torch.float32
        )

        pipe_kwargs = {
            "tokenizer": None,
            "safety_checker": None,
            "feature_extractor": None,
            "requires_safety_checker": False,
            "torch_dtype": self.weights_dtype,
        }
        self.pipe = StableDiffusionPipeline.from_pretrained(
            self.cfg.pretrained_model_name_or_path,
            **pipe_kwargs,
        ).to(self.device)

        if self.cfg.enable_memory_efficient_attention:
            if parse_version(torch.__version__) >= parse_version("2"):
                threestudio.info(
                    "PyTorch2.0 uses memory efficient attention by default."
                )
            elif not is_xformers_available():
                threestudio.warn(
                    "xformers is not available, memory efficient attention is not enabled."
                )
            else:
                self.pipe.enable_xformers_memory_efficient_attention()

        if self.cfg.enable_sequential_cpu_offload:
            self.pipe.enable_sequential_cpu_offload()

        if self.cfg.enable_attention_slicing:
            self.pipe.enable_attention_slicing(1)

        if self.cfg.enable_channels_last_format:
            self.pipe.unet.to(memory_format=torch.channels_last)

        del self.pipe.text_encoder
        cleanup()

        # Create model
        self.vae = self.pipe.vae.eval()
        self.unet = self.pipe.unet.eval()

        for p in self.vae.parameters():
            p.requires_grad_(False)
        for p in self.unet.parameters():
            p.requires_grad_(False)

        if self.cfg.token_merging:
            import tomesd
            tomesd.apply_patch(self.unet, **self.cfg.token_merging_params)

        if self.cfg.use_sjc:
            # score jacobian chaining use DDPM
            self.scheduler = DDPMScheduler.from_pretrained(
                self.cfg.pretrained_model_name_or_path,
                subfolder="scheduler",
                torch_dtype=self.weights_dtype,
                beta_start=0.00085,
                beta_end=0.0120,
                beta_schedule="scaled_linear",
            )
        else:
            self.scheduler = DDIMScheduler.from_pretrained(
                self.cfg.pretrained_model_name_or_path,
                subfolder="scheduler",
                torch_dtype=self.weights_dtype,
            )

        self.num_train_timesteps = self.scheduler.config.num_train_timesteps
        self.set_min_max_steps()  # set to default value

        self.alphas: Float[Tensor, "..."] = self.scheduler.alphas_cumprod.to(
            self.device
        )
        if self.cfg.use_sjc:
            # score jacobian chaining need mu
            self.us: Float[Tensor, "..."] = torch.sqrt((1 - self.alphas) / self.alphas)

        self.grad_clip_val: Optional[float] = None

        threestudio.info(f"Loaded Stable Diffusion!")
        
        # Ours
        self.inverse_scheduler = DDIMInverseScheduler.from_pretrained(
            self.cfg.pretrained_model_name_or_path,
            subfolder="scheduler",
            torch_dtype=self.weights_dtype
        )
        self.inverse_scheduler.set_timesteps(self.cfg.ddim_inverse_n_steps, device=self.device)
        self.inverse_scheduler.alphas_cumprod = self.inverse_scheduler.alphas_cumprod.to(device=self.device)
        if self.cfg.reset_t_every < 0:
            self.cfg.reset_t_every = self.cfg.trainer_max_steps
        self.fixed_noise_init = False

    @torch.cuda.amp.autocast(enabled=False)
    def set_min_max_steps(self, min_step_percent=0.02, max_step_percent=0.98):
        self.min_step = int(self.num_train_timesteps * min_step_percent)
        self.max_step = int(self.num_train_timesteps * max_step_percent)

    @torch.cuda.amp.autocast(enabled=False)
    def forward_unet(
        self,
        latents: Float[Tensor, "..."],
        t: Float[Tensor, "..."],
        encoder_hidden_states: Float[Tensor, "..."],
    ) -> Float[Tensor, "..."]:
        input_dtype = latents.dtype
        return self.unet(
            latents.to(self.weights_dtype),
            t.to(self.weights_dtype),
            encoder_hidden_states=encoder_hidden_states.to(self.weights_dtype),
        ).sample.to(input_dtype)

    @torch.cuda.amp.autocast(enabled=False)
    def encode_images(
        self, imgs: Float[Tensor, "B 3 512 512"]
    ) -> Float[Tensor, "B 4 64 64"]:
        input_dtype = imgs.dtype
        imgs = imgs * 2.0 - 1.0
        posterior = self.vae.encode(imgs.to(self.weights_dtype)).latent_dist
        latents = posterior.sample() * self.vae.config.scaling_factor
        return latents.to(input_dtype)

    @torch.cuda.amp.autocast(enabled=False)
    def decode_latents(
        self,
        latents: Float[Tensor, "B 4 H W"],
        latent_height: int = 64,
        latent_width: int = 64,
    ) -> Float[Tensor, "B 3 512 512"]:
        input_dtype = latents.dtype
        latents = F.interpolate(
            latents, (latent_height, latent_width), mode="bilinear", align_corners=False
        )
        latents = 1 / self.vae.config.scaling_factor * latents
        image = self.vae.decode(latents.to(self.weights_dtype)).sample
        image = (image * 0.5 + 0.5).clamp(0, 1)
        return image.to(input_dtype)
    

    def regress_fixed_point_iteration(self, img, t, prompt_utils, elevation, azimuth, camera_distances, n_iters=10):
        with torch.no_grad():
            kappa = torch.randn_like(img)
            for i in range(n_iters):
                noisy_img = self.scheduler.add_noise(img, kappa, t)
                kappa, _, _ = self.get_noise_pred(noisy_img, t, prompt_utils, elevation, azimuth, camera_distances)
            return kappa

    def regress_sgd_opt(self, img, t, prompt_utils, elevation, azimuth, camera_distances, n_iters=10):
        kappa = torch.randn_like(img).requires_grad_(True)
        optimizer = torch.optim.SGD([kappa], lr=1e-1)
        for i in range(n_iters):
            optimizer.zero_grad()

            noisy_img = self.scheduler.add_noise(img, kappa, t)
            noise_pred, _, _ = self.get_noise_pred(noisy_img, t, prompt_utils, elevation, azimuth, camera_distances)
            
            loss = F.mse_loss(noise_pred.detach().clone(), kappa)
            loss.backward()
            optimizer.step()
        return kappa.detach().requires_grad_(False)

    def compute_grad_sds(
        self,
        latents: Float[Tensor, "B 4 64 64"],
        image: Float[Tensor, "B 3 512 512"],
        t: Int[Tensor, "B"],
        prompt_utils: PromptProcessorOutput,
        elevation: Float[Tensor, "B"],
        azimuth: Float[Tensor, "B"],
        camera_distances: Float[Tensor, "B"],
    ):
        batch_size = elevation.shape[0]
        
        with torch.no_grad():
            if self.cfg.noise_strategy == "rand":
                # print("using rand noise strategy")
                noise = torch.randn_like(latents)
                latents_noisy = self.scheduler.add_noise(latents, noise, t)
            elif self.cfg.noise_strategy == "randf":
                if not self.fixed_noise_init:
                    self.fixed_noise = torch.randn_like(latents)
                    self.fixed_noise_init = True
                latents_noisy = self.scheduler.add_noise(latents, self.fixed_noise, t)
            elif self.cfg.noise_strategy == "fixedpoint":
                noise = self.regress_fixed_point_iteration(latents, t, prompt_utils, elevation, azimuth, camera_distances)
                latents_noisy = self.scheduler.add_noise(latents, noise, t)
            elif self.cfg.noise_strategy == "sgd":
                noise = self.regress_sgd_opt(latents, t, prompt_utils, elevation, azimuth, camera_distances)
                latents_noisy = self.scheduler.add_noise(latents, noise, t)
            else:
                ## Ours
                # print("using ours noise strategy")
                latents_noisy = self.get_noisy_x(t, latents, prompt_utils, elevation, azimuth, camera_distances)
            
            noise_pred, text_embeddings, neg_guidance_weights = self.get_noise_pred(latents_noisy, t, prompt_utils, elevation, azimuth, camera_distances)

        if self.cfg.use_sds_scaling:
            with torch.no_grad():
                if self.cfg.weighting_strategy == "sds":
                    # w(t), sigma_t^2
                    w = (1 - self.alphas[t]).view(-1, 1, 1, 1)
                elif self.cfg.weighting_strategy == "uniform":
                    w = 1
                elif self.cfg.weighting_strategy == "fantasia3d":
                    w = (self.alphas[t] ** 0.5 * (1 - self.alphas[t])).view(-1, 1, 1, 1)
                else:
                    raise ValueError(
                        f"Unknown weighting strategy: {self.cfg.weighting_strategy}"
                    )
                noise = self.get_noise_from_target(latents, latents_noisy, t)
                grad = w * (noise_pred - noise)
                grad = torch.nan_to_num(grad)
                if self.grad_clip_val is not None:
                    grad = grad.clamp(-self.grad_clip_val, self.grad_clip_val)
                latents_denoised = (latents - grad).detach()
        else:
            latents_denoised = self.get_x0(latents_noisy, noise_pred, t).detach()

        guidance_eval_utils = {
            "use_perp_neg": prompt_utils.use_perp_neg,
            "neg_guidance_weights": neg_guidance_weights,
            "text_embeddings": text_embeddings,
            "t_orig": t,
            "latents_noisy": latents_noisy,
            "noise_pred": noise_pred,
            "target_x0": latents_denoised,
        }
        
        # Last step - update target for the dual trajectory
        return latents_denoised.detach(), guidance_eval_utils, latents_noisy
    
    def make_inverse_step(self, scheduler, model_output, sample, timestep, prev_timestep):
        alpha_prod_t = scheduler.alphas_cumprod[timestep] if timestep >= 0 else scheduler.initial_alpha_cumprod
        alpha_prod_t_prev = scheduler.alphas_cumprod[prev_timestep]

        beta_prod_t = 1 - alpha_prod_t

        # 3. compute predicted original sample from predicted noise also called
        # "predicted x_0" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        if scheduler.config.prediction_type == "epsilon":
            pred_original_sample = (sample - beta_prod_t ** (0.5) * model_output) / alpha_prod_t ** (0.5)
            pred_epsilon = model_output
        elif scheduler.config.prediction_type == "sample":
            pred_original_sample = model_output
            pred_epsilon = (sample - alpha_prod_t ** (0.5) * pred_original_sample) / beta_prod_t ** (0.5)
        elif scheduler.config.prediction_type == "v_prediction":
            pred_original_sample = (alpha_prod_t**0.5) * sample - (beta_prod_t**0.5) * model_output
            pred_epsilon = (alpha_prod_t**0.5) * model_output + (beta_prod_t**0.5) * sample
        else:
            raise ValueError(
                f"prediction_type given as {scheduler.config.prediction_type} must be one of `epsilon`, `sample`, or"
                " `v_prediction`"
            )

        # 4. Clip or threshold "predicted x_0"
        if scheduler.config.clip_sample:
            pred_original_sample = pred_original_sample.clamp(
                -scheduler.config.clip_sample_range, scheduler.config.clip_sample_range
            )

        # 5. compute "direction pointing to x_t" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        pred_sample_direction = (1 - alpha_prod_t_prev) ** (0.5) * pred_epsilon

        # 6. compute x_t without "random noise" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        prev_sample = alpha_prod_t_prev ** (0.5) * pred_original_sample + pred_sample_direction
        return prev_sample
        
        
    def get_inverse_timesteps(self, cur_t):
        
        if self.cfg.use_legacy_scheduler:
            timesteps = self.inverse_scheduler.timesteps[self.inverse_scheduler.timesteps < cur_t]
            return timesteps
        # Set the timesteps to properly match the target t
        num_inference_steps = self.cfg.ddim_inverse_n_steps
        num_train_timesteps = self.inverse_scheduler.config.num_train_timesteps
      
        # NOTE: t is shared within batch
        n_steps_to_cur_t = max(1, int(num_inference_steps * torch.max(cur_t).item() / num_train_timesteps))
        self.inverse_scheduler.num_inference_steps = n_steps_to_cur_t - 1 # need to substract 1 to make sure the prev_t is matching exactly is 
        
        step_ratio = (cur_t / n_steps_to_cur_t)[0].item()
        if self.inverse_scheduler.config.timestep_spacing == "leading":
            timesteps = (np.arange(0, n_steps_to_cur_t) * step_ratio).round().copy().astype(np.int64)
            timesteps += self.inverse_scheduler.config.steps_offset
        elif self.inverse_scheduler.config.timestep_spacing == "trailing":
            timesteps = np.round(np.arange(cur_t, 0, -step_ratio)[::-1]).astype(np.int64)
            timesteps -= 1
    
        timesteps = np.insert(timesteps, 0, -int(step_ratio)) # Seems important for the first step
        timesteps = np.append(timesteps, cur_t.item())
        timesteps = torch.from_numpy(timesteps).to(self.device)
        return timesteps
    
    def inference_2d(self, start_latents, cur_t, scheduler, prompt_utils, elevation, azimuth, camera_distances):
        latents = start_latents.clone()
        B = start_latents.shape[0]
        
        timesteps = self.get_inverse_timesteps(cur_t)
        
        inversion_eta = self.cfg.inversion_eta 
        if self.cfg.linear_inversion_entropy:
            inversion_eta = ( torch.abs(azimuth / 180) ** self.cfg.inversion_entropy_power) * inversion_eta
    
        # print(cur_t, timesteps)
        with torch.no_grad():
            if self.cfg.use_legacy_scheduler:
                for t in timesteps:
                    noise_pred, _, _ =  self.get_noise_pred(latents, t.repeat([B]), prompt_utils, elevation, azimuth, camera_distances,
                                                         guidance_scale=self.cfg.inverse_guidance_scale)
                    latents = scheduler.step(noise_pred, t, latents).prev_sample
                    
                    prev_t = t + self.inverse_scheduler.config.num_train_timesteps // self.inverse_scheduler.num_inference_steps
                    variance = self.scheduler._get_variance(prev_t, t) ** (0.5)
                    latents += inversion_eta * torch.randn_like(latents) * variance
            else:
                for t, prev_t in zip(timesteps[:-1], timesteps[1:]):
                    noise_pred, _, _ =  self.get_noise_pred(latents, t.repeat([B]), prompt_utils, elevation, azimuth, camera_distances,
                                                         guidance_scale=self.cfg.inverse_guidance_scale)
                    latents = self.make_inverse_step(scheduler, noise_pred, latents, t, prev_t)
                    
                    variance = self.scheduler._get_variance(prev_t, t) ** (0.5)
                    latents += inversion_eta * torch.randn_like(latents) * variance
        return latents
    
    # TODO: test this
    def get_noise_from_target(self, target, cur_xt, t):
        alpha_prod_t = self.scheduler.alphas_cumprod[t]
        beta_prod_t = 1 - alpha_prod_t
        noise = (cur_xt - target * alpha_prod_t ** (0.5)) / (beta_prod_t ** (0.5))
        return noise
    
    def get_x0(self, original_samples, noise_pred, t):
        step_results = self.scheduler.step(noise_pred, t[0], original_samples, return_dict=True)
        if "pred_original_sample" in step_results:
            return step_results["pred_original_sample"]
        elif "denoised" in step_results:
            return step_results["denoised"]
        raise ValueError("Looks like the scheduler does not compute x0")
    
    def get_noisy_x(self, t, latent, prompt_utils, elevation, azimuth, camera_distances):
        with torch.no_grad():
            if t < self.cfg.inversion_t_threshold:
                return self.inference_2d(
                    latent,
                    t,
                    self.inverse_scheduler,
                    prompt_utils,
                    elevation, azimuth, camera_distances
                )
            else:
                noise = torch.randn_like(latent)
                return self.scheduler.add_noise(latent, noise, t)

    def __call__(
        self,
        rgb: Float[Tensor, "B H W C"],
        prompt_utils: PromptProcessorOutput,
        elevation: Float[Tensor, "B"],
        azimuth: Float[Tensor, "B"],
        camera_distances: Float[Tensor, "B"],
        rgb_as_latents=False,
        guidance_eval=False,
        test_call=False,
        **kwargs,
    ):
        # TODO: Find a better place for it
        self.scheduler.alphas_cumprod = self.scheduler.alphas_cumprod.to(device=self.device)
        self.scheduler.set_timesteps(self.cfg.n_ddim_timesteps, device=self.device)

        batch_size = rgb.shape[0]

        rgb_BCHW = rgb.permute(0, 3, 1, 2)
        latents: Float[Tensor, "B 4 64 64"]
        rgb_BCHW_512 = F.interpolate(
            rgb_BCHW, (512, 512), mode="bilinear", align_corners=False
        )
        if rgb_as_latents:
            latents = F.interpolate(
                rgb_BCHW, (64, 64), mode="bilinear", align_corners=False
            )
        else:
            # encode image into latents with vae
            latents = self.encode_images(rgb_BCHW_512)

        # timestep ~ U(0.02, 0.98) to avoid very high/low noise level
        t = torch.randint(
            self.min_step,
            self.max_step + 1,
            [batch_size],
            dtype=torch.long,
            device=self.device,
        )
       
        target_img, guidance_eval_utils, noisy_image = self.compute_grad_sds(
            latents,
            rgb_BCHW_512,
            t,
            prompt_utils,
            elevation,
            azimuth,
            camera_distances
        )
        
        if test_call:
            return target_img, noisy_image

        loss_sds = 0.5 * F.mse_loss(latents, target_img.detach(), reduction=self.cfg.loss_reduction) / batch_size

        guidance_out = {
            "loss_sds": loss_sds,
            "grad_norm": (latents - target_img).norm(),
            "min_step": self.min_step,
            "max_step": self.max_step,
            "target_x0": guidance_eval_utils["target_x0"]
        }

        if self.cfg.use_img_loss:
            grad_img = torch.nan_to_num(grad_img)
            if self.grad_clip_val is not None:
                grad_img = grad_img.clamp(-self.grad_clip_val, self.grad_clip_val)
            target_img = (rgb_BCHW_512 - grad_img).detach()
            loss_sds_img = (
                0.5 * F.mse_loss(rgb_BCHW_512, target_img, reduction=self.cfg.loss_reduction) / batch_size
            )
            guidance_out["loss_sds_img"] = loss_sds_img

        if guidance_eval:
            guidance_eval_out = self.guidance_eval(**guidance_eval_utils)
            texts = []
            for n, e, a, c in zip(
                guidance_eval_out["noise_levels"], elevation, azimuth, camera_distances
            ):
                texts.append(
                    f"n{n:.02f}\ne{e.item():.01f}\na{a.item():.01f}\nc{c.item():.02f}"
                )
            guidance_eval_out.update({"texts": texts})
            guidance_out.update({"eval": guidance_eval_out})

        return guidance_out

    @torch.cuda.amp.autocast(enabled=False)
    @torch.no_grad()
    def get_noise_pred(
        self,
        latents_noisy,
        t,
        prompt_utils,
        elevation, azimuth, camera_distances,
        neg_guidance_weights=None,
        guidance_scale=None,
    ):
        if guidance_scale is None:
            guidance_scale = self.cfg.guidance_scale
        guidance_scale = guidance_scale + torch.abs(azimuth / 180) * (self.cfg.back_cfg_scale - 1) * guidance_scale
        
        batch_size = latents_noisy.shape[0]

        if prompt_utils.use_perp_neg:
            (
                text_embeddings,
                neg_guidance_weights,
            ) = prompt_utils.get_text_embeddings_perp_neg(
                elevation, azimuth, camera_distances, self.cfg.view_dependent_prompting
            )
            
            # pred noise
            latent_model_input = torch.cat([latents_noisy] * 4, dim=0)
            noise_pred = self.forward_unet(
                latent_model_input,
                torch.cat([t.reshape(1)] * 4).to(self.device),
                encoder_hidden_states=text_embeddings,
            )  # (4B, 3, 64, 64)

            noise_pred_text = noise_pred[:batch_size]
            noise_pred_uncond = noise_pred[batch_size : batch_size * 2]
            noise_pred_neg = noise_pred[batch_size * 2 :]

            e_pos = noise_pred_text - noise_pred_uncond
            accum_grad = 0
            n_negative_prompts = neg_guidance_weights.shape[-1]
            for i in range(n_negative_prompts):
                e_i_neg = noise_pred_neg[i::n_negative_prompts] - noise_pred_uncond
                accum_grad += neg_guidance_weights[:, i].view(
                    -1, 1, 1, 1
                ) * perpendicular_component(e_i_neg, e_pos)

            noise_pred = noise_pred_uncond + guidance_scale * (
                e_pos + accum_grad
            )
        elif self.cfg.prompt_augmentation_cfg_scale > 1.0 + 1e-6:
            neg_guidance_weights = None
            text_embeddings, direction_idx = prompt_utils.get_text_embeddings_view_aug(
                elevation, azimuth, camera_distances, self.cfg.view_dependent_prompting
            )
        
            # Disable front CFG augmentation
            pormpt_cfg_aug = self.cfg.prompt_augmentation_cfg_scale * torch.ones_like(direction_idx)
            pormpt_cfg_aug[direction_idx == 0] = 1.
            
            # pred noise
            latent_model_input = torch.cat([latents_noisy] * 3, dim=0)
            noise_pred = self.forward_unet(
                latent_model_input,
                torch.cat([t.reshape(-1)] * 3).to(self.device),
                encoder_hidden_states=text_embeddings,
            )
            noise_pred_orig, noise_pred_vd, noise_pred_uncond = noise_pred.chunk(3)
            noise_pred = noise_pred_uncond + guidance_scale * (
                noise_pred_orig - noise_pred_uncond
            ) + pormpt_cfg_aug * guidance_scale * (
                noise_pred_vd - noise_pred_orig
            )
        else:
            neg_guidance_weights = None
            text_embeddings = prompt_utils.get_text_embeddings(
                elevation, azimuth, camera_distances, self.cfg.view_dependent_prompting
            )
            
            # pred noise
            latent_model_input = torch.cat([latents_noisy] * 2, dim=0)
            noise_pred = self.forward_unet(
                latent_model_input,
                torch.cat([t.reshape(-1)] * 2).to(self.device),
                encoder_hidden_states=text_embeddings,
            )
            # perform guidance (high scale from paper!)
            noise_pred_text, noise_pred_uncond = noise_pred.chunk(2)
            noise_pred = noise_pred_uncond + guidance_scale * (
                noise_pred_text - noise_pred_uncond
            )

        return noise_pred, text_embeddings, neg_guidance_weights

    @torch.cuda.amp.autocast(enabled=False)
    @torch.no_grad()
    def guidance_eval(
        self,
        t_orig,
        text_embeddings,
        latents_noisy,
        noise_pred,
        azimuth, 
        use_perp_neg=False,
        neg_guidance_weights=None,
    ):
        # use only 50 timesteps, and find nearest of those to t
        self.scheduler.set_timesteps(50)
        self.scheduler.timesteps_gpu = self.scheduler.timesteps.to(self.device)
        bs = (
            min(self.cfg.max_items_eval, latents_noisy.shape[0])
            if self.cfg.max_items_eval > 0
            else latents_noisy.shape[0]
        )  # batch size
        large_enough_idxs = self.scheduler.timesteps_gpu.expand([bs, -1]) > t_orig[
            :bs
        ].unsqueeze(
            -1
        )  # sized [bs,50] > [bs,1]
        idxs = torch.min(large_enough_idxs, dim=1)[1]
        t = self.scheduler.timesteps_gpu[idxs]

        fracs = list((t / self.scheduler.config.num_train_timesteps).cpu().numpy())
        imgs_noisy = self.decode_latents(latents_noisy[:bs]).permute(0, 2, 3, 1)

        # get prev latent
        latents_1step = []
        pred_1orig = []
        for b in range(bs):
            step_output = self.scheduler.step(
                noise_pred[b : b + 1], t[b], latents_noisy[b : b + 1], eta=1
            )
            latents_1step.append(step_output["prev_sample"])
            pred_1orig.append(step_output["pred_original_sample"])
        latents_1step = torch.cat(latents_1step)
        pred_1orig = torch.cat(pred_1orig)
        imgs_1step = self.decode_latents(latents_1step).permute(0, 2, 3, 1)
        imgs_1orig = self.decode_latents(pred_1orig).permute(0, 2, 3, 1)

        latents_final = []
        for b, i in enumerate(idxs):
            latents = latents_1step[b : b + 1]
            text_emb = (
                text_embeddings[
                    [b, b + len(idxs), b + 2 * len(idxs), b + 3 * len(idxs)], ...
                ]
                if use_perp_neg
                else text_embeddings[[b, b + len(idxs)], ...]
            )
            neg_guid = neg_guidance_weights[b : b + 1] if use_perp_neg else None
            for t in tqdm(self.scheduler.timesteps[i + 1 :], leave=False):
                # pred noise
                noise_pred, text_emb = self.get_noise_pred(
                    latents, t, prompt_utils, azimuth
                )
                # get prev latent
                latents = self.scheduler.step(noise_pred, t, latents, eta=1)[
                    "prev_sample"
                ]
            latents_final.append(latents)

        latents_final = torch.cat(latents_final)
        imgs_final = self.decode_latents(latents_final).permute(0, 2, 3, 1)

        return {
            "bs": bs,
            "noise_levels": fracs,
            "imgs_noisy": imgs_noisy,
            "imgs_1step": imgs_1step,
            "imgs_1orig": imgs_1orig,
            "imgs_final": imgs_final,
        }

    def update_step(self, epoch: int, global_step: int, on_load_weights: bool = False):
        # clip grad for stable training as demonstrated in
        # Debiasing Scores and Prompts of 2D Diffusion for Robust Text-to-3D Generation
        # http://arxiv.org/abs/2303.15413
        
        max_steps = self.cfg.trainer_max_steps
        
        max_steps = min(self.cfg.reset_t_every, self.cfg.trainer_max_steps - self.cfg.reset_t_every * (global_step // self.cfg.reset_t_every))
        if max_steps == 0:
            max_steps = self.cfg.reset_t_every
        global_step = global_step % self.cfg.reset_t_every
    
        if self.cfg.grad_clip is not None:
            self.grad_clip_val = C(self.cfg.grad_clip, epoch, global_step)

        if self.cfg.sqrt_anneal:
            percentage = (
                float(global_step) / max_steps
            ) ** self.cfg.t_map_power  # progress percentage
            if type(self.cfg.max_step_percent) not in [float, int]:
                max_step_percent = self.cfg.max_step_percent[1]
            else:
                max_step_percent = self.cfg.max_step_percent
            curr_percent = (
                max_step_percent - C(self.cfg.min_step_percent, epoch, global_step)
            ) * (1 - percentage) + C(self.cfg.min_step_percent, epoch, global_step)
            self.set_min_max_steps(
                min_step_percent=curr_percent,
                max_step_percent=curr_percent,
            )
        else:
            self.set_min_max_steps(
                min_step_percent=C(self.cfg.min_step_percent, epoch, global_step),
                max_step_percent=C(self.cfg.max_step_percent, epoch, global_step),
            )
            
        # print(global_step, max_steps, self.min_step, self.max_step)